@@ -1,20 +1,11 @@
 import numpy as np
 import warnings
 from collections import OrderedDict, namedtuple
-from scipy.integrate import odeint
+from scipy.integrate import odeint, solve_ivp
 from scipy.optimize import minimize
-try:
-    from numba import jit
-except ImportError:
-    warnings.warn('Numba not found - Gillespie algorithm calculations'
-                  ' may be inefficient')
+from numba import jit
 
-    def jit(*args, **kwargs):
-        def fakedec(f):
-            return f
-        return fakedec
-
-FitResult = namedtuple('FitResult', ['C', 'y0', 'R2', 'RMSRE', 'success'])
+FitResult = namedtuple('FitResult', ['C', 'y0', 'R2', 'MRPD', 'success'])
 
 
 @jit(nopython=True)
@@ -129,16 +120,16 @@ class CModel(object):
         of three types:
 
         * CONSTANT
-            A constant rate of growth for a certain state of the 
+            A constant rate of growth for a certain state of the
             model.
         * LINEAR
             A rate of growth that is linearly proportional to the population
             of the same or another state.
         * QUADRATIC
-            A rate of growth that is proportional to the product of the 
+            A rate of growth that is proportional to the product of the
             populations of two states.
 
-        The coupling is defined by a descriptor string. This can have up to 
+        The coupling is defined by a descriptor string. This can have up to
         four states:
 
         "S1*S2:S3=>S4"
@@ -160,19 +151,19 @@ class CModel(object):
 
         "S1:=>S4"
 
-        results in 
+        results in
 
         dS4/dt = C*S1
 
         "S3=>S4"
 
-        results in 
+        results in
 
         dS3/dt = -C
 
         dS4/dt = C
 
-        and so on. 
+        and so on.
 
         Arguments:
             descr {str} -- Descriptor string for the coupling.
@@ -242,12 +233,36 @@ class CModel(object):
                                                 for s in (s1, s2, s3, s4)])
             name = descr
 
+        if name in self._couplings:
+            raise ValueError('Coupling {0} already exists '
+                             'for model'.format(name))
+
         self._couplings[name] = (descr, C)
 
         self._cdata['C'] = np.concatenate([self._cdata['C'], [C]])
         self._cdata['i'] = np.concatenate(
             [self._cdata['i'], [[i1, i2, i3, i4]]], axis=0)
 
+    def edit_coupling_rate(self, name, C):
+        """Change the coupling rate for an existing coupling
+
+        Change the coupling rate for an existing coupling
+
+        Arguments:
+            name {str} -- Name of the coupling
+            C {number} -- New value
+        """
+
+        names = list(self._couplings.keys())
+        try:
+            i = names.index(name)
+        except ValueError:
+            raise ValueError('No coupling with name {0} exists'.format(name))
+
+        descr, _ = self._couplings[name]
+        self._couplings[name] = (descr, C)
+        self._cdata['C'][i] = C
+
     def dy_dt(self, y):
         """Time derivative from a given state
 
@@ -355,11 +370,11 @@ class CModel(object):
 
         return d2ydtdy0
 
-    def integrate(self, t, y0, use_gradient=False):
+    def integrate(self, t, y0, use_gradient=False, events=None):
         """Integrate the ODEs of the model
 
         Carry out an integration of the system of ODEs representing
-        the model using the scipy.integrate.odeint method.
+        the model using the scipy.integrate.solve_ivp method.
 
         Arguments:
             t {np.ndarray} -- Array of times for the integration
@@ -369,6 +384,11 @@ class CModel(object):
             use_gradient {bool} -- If True, also compute the
             gradient of the curves w.r.t parameters and initial conditions
             (default: {False})
+            events {[fun(t, y)]} -- List of functions of t, y (where y is the
+            array of states, or states + gradients if use_gradient is True)
+            at which special events happen and are recorded, or can terminate
+            integration. See scipy's documentation for solve_ivp for further
+            details (default: {None})
 
         Returns:
             OrderedDict -- Dictionary of the computed trajectories, with the
@@ -376,22 +396,28 @@ class CModel(object):
                 - y: trajectory of the main populations
                 - dy/d([coupling name]): derivatives wrt coupling constants
                 - dy/d([state name]0): derivatives wrt starting conditions
+                - t: times of integration. Usually the entire model.t, but can
+                be less in presence of terminating events.
+                - t_events: times of events.
 
-            The latter two are present only if use_gradient = True.
+            The derivatives are present only if use_gradient = True. The event
+            times are present only if events is not None.
         """
 
         if not use_gradient:
-            def ode(y, t):
+            def ode(t, y):
                 return self.dy_dt(y)
 
-            traj = odeint(ode, y0, t)
+            sol = solve_ivp(ode, [t[0], t[-1]], y0, t_eval=t,
+                            events=events)
+            traj = sol.y.T
 
             ans = OrderedDict({'y': traj})
 
         else:
             N2 = self._N**2
 
-            def ode(y, t):
+            def ode(t, y):
                 dydt = y*0.
                 dydt[:self._N] = self.dy_dt(y[:self._N])
                 dydt[self._N:-N2] = self.d2y_dtdC(y[:self._N],
@@ -404,7 +430,9 @@ class CModel(object):
             y0 = np.concatenate([y0, np.zeros(nC*self._N),
                                  np.eye(self._N).reshape((-1,))])
 
-            traj = odeint(ode, y0, t)
+            sol = solve_ivp(ode, [t[0], t[-1]], y0, t_eval=t,
+                            events=events)
+            traj = sol.y.T
 
             ans = OrderedDict({'y': traj[:, :self._N]})
 
@@ -418,6 +446,10 @@ class CModel(object):
                                                       i*self._N+i0:(i+1) *
                                                       self._N+i0]
 
+        ans['t'] = sol.t
+        if events is not None:
+            ans['t_events'] = sol.t_events
+
         return ans
 
     def gillespie(self, tmax, y0, samples=1000):
@@ -430,13 +462,13 @@ class CModel(object):
             y0 {np.ndarray} -- Starting state
 
         Keyword Arguments:
-            samples {number} -- Number of trajectories to sample 
+            samples {number} -- Number of trajectories to sample
                                 (default: {1000})
 
         Returns:
             OrderedDict -- Dictionary of the computed trajectories, with the
             following keys:
-                - t: times at which the state is computed, for each 
+                - t: times at which the state is computed, for each
                      trajectory
                 - y: trajectories of the main populations
         """
@@ -457,7 +489,7 @@ class CModel(object):
 
         return ans
 
-    def fit(self, data, steps=1000):
+    def fit(self, data, steps=1000, constraints={}):
         """Fit the model to data
 
         Fit the model's parameters to a given data set, minimising the
@@ -466,7 +498,7 @@ class CModel(object):
 
         Arguments:
             data {list} -- A list of the data points available. Can be a list
-            of dictionaries or arrays. If it's dictionaries, they must all 
+            of dictionaries or arrays. If it's dictionaries, they must all
             have a member `t' (for the time of the point) and then can have
             members with the names of various compartments of the model. If
             it's arrays, then they must have N+1 elements for N compartments;
@@ -476,6 +508,11 @@ class CModel(object):
         Keyword Arguments:
             steps {number} -- Number of integration steps to split the time
             interval into (default: {1000})
+            constraints {dict} -- Dictionary whose keys must be valid names of
+            coupling constants or X0 with X a valid name of a state, and whose
+            values must be numbers.
+            These constants, or the initial conditions for those states,
+            will be fixed to those values and not allowed to change.
 
         Returns:
             FitResult - Named tuple containing the following members:
@@ -483,7 +520,7 @@ class CModel(object):
                 - y0: fitted optimal starting state
                 - R2: array of R squared values for goodness of fit, one
                 for each curve (can be R2 > 1 due to this)
-                - RMSRE: Root Mean Square Relative Error, total
+                - MRPD: Mean Relative Percent Difference, total
                 - success: if True, the fitting has converged to a final value.
 
         Raises:
@@ -522,47 +559,71 @@ class CModel(object):
         # Now sort, and identify the data points
         tsort = np.argsort(t)
         t = t[tsort]
+        # Make it unique
+        t, uniq_inv = np.unique(t, return_inverse=True)
         data_i = np.where(tsort - steps >= 0)[0]
 
         C = self._cdata['C']
         nC = len(C)
 
+        # Get constraints mask
+        mask = np.ones(nC+self._N)
+        y0 = np.ones(self._N)*np.nan
+        for i, name in enumerate(self._couplings.keys()):
+            if name in constraints:
+                self.edit_coupling_rate(name, constraints[name])
+                mask[i] = 0
+        for i, name in enumerate(self._states):
+            if name + '0' in constraints:
+                y0[i] = constraints[name + '0']
+                mask[nC+i] = 0
+
         # Cost function
+
         def costf(x):
             self._cdata['C'] = x[:nC]
             y0 = x[nC:]
             traj = self.integrate(t, y0, use_gradient=True)
-            traj = np.array(list(traj.values()))
+            traj = np.array(list(traj.values())[:-1])
+            # Rebuild original array
+            traj = traj[:, uniq_inv, :]
 
             # Gradient?
             yref = traj[:, data_i, :]
             err = (data[:, 1:]-yref[0])
             err = np.where(np.isnan(data[:, 1:]), 0, err)
-            return np.sum(err**2)
+            cost = np.sum(err**2)
+            return cost
 
         # Gradient of cost function
         def costfgrad(x):
             self._cdata['C'] = x[:nC]
             y0 = x[nC:]
             traj = self.integrate(t, y0, use_gradient=True)
-            traj = np.array(list(traj.values()))
+            traj = np.array(list(traj.values())[:-1])
+            # Rebuild original array
+            traj = traj[:, uniq_inv, :]
 
             # Gradient?
             yref = traj[:, data_i, :]
             err = (yref[0]-data[:, 1:])
             err = np.where(np.isnan(data[:, 1:]), 0, err)
-            return np.sum(2*err[None]*yref[1:, :, :], axis=(1, 2))
+            return np.sum(2*err[None]*yref[1:, :, :]*mask[:, None, None],
+                          axis=(1, 2))
 
         x0 = data[0, 1:]
         x0 = np.where(np.isnan(x0), 0, x0)
+        x0 = np.where(np.isnan(y0), x0, y0)         # Apply constraints
         x0 = np.concatenate([self._cdata['C'], x0])
         sol = minimize(costf, x0, jac=costfgrad)
 
         x = sol.x
-        self._cdata['C'] = x[:nC]
+        for i, name in enumerate(self._couplings.keys()):
+            self.edit_coupling_rate(name, x[i])
         y0 = x[nC:]
 
         traj = self.integrate(t, y0)['y']
+        traj = traj[uniq_inv, :]
 
         # R2?
         yref = traj[data_i, :]
@@ -578,23 +639,26 @@ class CModel(object):
 
         fitC = {c: x[i] for i, c in enumerate(self._couplings.keys())}
 
-        RMSRE = (np.sum(((yref*dataMask-data[:, 1:])/yref)**2)
-                 / np.sum(dataN))**0.5
+        avgAbs = (np.abs(yref*dataMask)+np.abs(data[:, 1:]))/2.0
+        avgAbs = np.where(avgAbs > 0, avgAbs, np.inf)
 
-        ans = FitResult(fitC, y0, SSreg/SStot, RMSRE, sol.success)
+        MRPD = (np.sum(np.abs(yref*dataMask-data[:, 1:])/ avgAbs)
+                / np.sum(dataN))
+        
+        ans = FitResult(fitC, y0, SSreg/SStot, MRPD, sol.success)
 
         return ans
 
     @staticmethod
     def make_SIR(beta=0.3, gamma=0.2):
         """Make a SIR model
-        
+
         Constructor for an epidemic SIR model
-        
+
         Keyword Arguments:
             beta {number} -- Infection rate (default: {0.3})
             gamma {number} -- Recovery rate (default: {0.2})
-        
+
         Returns:
             CModel -- A CModel object describing a SIR model
         """
@@ -608,18 +672,17 @@ class CModel(object):
     @staticmethod
     def make_SIS(beta=0.3, gamma=0.2):
         """Make a SIS model
-        
+
         Constructor for an epidemic SIS model
-        
+
         Keyword Arguments:
             beta {number} -- Infection rate (default: {0.3})
             gamma {number} -- Recovery rate (default: {0.2})
-        
+
         Returns:
             CModel -- A CModel object describing a SIS model
         """
 
-
         sir = CModel('SI')
         sir.set_coupling_rate('S*I', beta, name='beta')
         sir.set_coupling_rate('I:I=>S', gamma, name='gamma')
@@ -629,20 +692,19 @@ class CModel(object):
     @staticmethod
     def make_LotkaVolterra(alpha=2.0/3.0, beta=4.0/3.0, gamma=1, delta=1):
         """Make a Lotka-Volterra model
-        
+
         Constructor for a Lotka-Volterra model of predation
-        
+
         Keyword Arguments:
             alpha {number} -- Rate of prey growth (default: {0.666})
             beta {number} -- Rate of prey killing (default: {1.333})
             gamma {number} -- Rate of predator eating (default: {1})
             delta {number} -- Rate of predator death (default: {1})
-        
+
         Returns:
             CModel -- A CModel object describing a Lotka-Volterra model
         """
 
-
         lv = CModel('Pp')
         lv.set_coupling_rate('p:=>p', alpha, 'alpha')
         lv.set_coupling_rate('p*P:p=>', beta, 'beta')
